#include <sys/types.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <limits.h>
#include "conf.h"

struct odpath_t *root;

struct odpath_t *find_path(const char *path, pathtype_t type) {
char seek[PATH_MAX + 3];
struct odpath_t *i;

	strcpy(seek, ":");
	strcat(seek, path);
	strcat(seek, ":");

	for (i = root; i != NULL; i = i->next)
		if (i->type == type && strstr(i->paths, seek))
			return i;

	return NULL;
}

#ifdef USE_ONENTERLEAVE
void display_file(const char *path) {
FILE *fp;

	if ((fp = fopen(path, "r"))) {
	char buffer[512];
	int len;

		while ((len = fread(buffer, 1, 512, fp)) > 0)
			fwrite(buffer, len, 1, stdout);
		fclose(fp);
	}
}

int exec_ok(const char *path) {
struct stat stat;

	return
		stat(path, &stat) == 0 && /* exists.. */
		s.st_mode & (S_IRUSR | S_IRGRP | S_IROTH) && /* ...readable... */
		s.st_uid == getuid() && /* ...owned by me... */
		!(s.st_mode & (S_IWGRP | S_IWOTH)); /* ...not group/world writeable */
}
#endif

struct odpath_t *find_enter(const char *path) {
	return find_path(path, PT_ENTER);
}

struct odpath_t *find_leave(const char *path) {
	return find_path(path, PT_LEAVE);
}

void usage(const char *msg) {
	printf(
		"%s\n"
		"usage: ondir [-V] <old-directory> [<new-directory>]\n"
		"\n"
		"  -V    Displays version information.\n"
		"\n"
		"  <old-directory> is the last working directory.\n"
		"\n"
		"  <new-directory> is the current working directory. If <new-directory> is\n"
		"  omitted, the current working directory is obtained via a call to getcwd().\n"
		"\n", msg
	);
	exit(1);
}

int main(int argc, const char **argv) {
#ifdef USE_ONENTERLEAVE
char onenter[PATH_MAX + 1];
#endif
char working[PATH_MAX + 1], cwd[PATH_MAX + 1];
int len, i;
const char *src = NULL, *dst = NULL, *home = NULL;

	if (argc < 2)
		usage("Not enough arguments");

	for (i = 1; i < argc; ++i) {
		if (!strcmp(argv[i], "-V")) {
			printf("This is OnDir version %s\n", VERSION);
			return 0;
		}

		if (argv[i][0] == '-')
			usage("Unknown argument");

		if (!src) src = argv[i];
		else
		if (!dst) dst = argv[i];
		else
			usage("You have already specified src and dst directories");
	}

	if (!dst) {
		getcwd(cwd, PATH_MAX);
		dst = cwd;
	}

	if (src[0] != '/' || dst[0] != '/')
		fatal("either the source or destination directory is not absolute");

	root = load_conf(GLOBAL_CONF, root);

	if ((home = getenv("HOME"))) {
		snprintf(working, PATH_MAX, "%s/.ondirrc", home);
		root = load_conf(working, root);
	}

	/*
		The basic algorithm is thus:

		1. Traverse up source path until it matches the lead of dest path 
		2. Traverse from the join to the end of dest path
	*/

	strcpy(working, src);
	len = strlen(working);

	/* Traverse up source path */
	while (strncmp(working, dst, len)) {
	struct odpath_t *p = find_leave(working);

		if (p) {
/*			printf("# Leaving %s\n", working);*/
			printf("%s", p->content);
		}
#ifdef USE_ONENTERLEAVE
		snprintf(onenter, PATH_MAX, "%s/.onleave", working);
		if (exec_ok(onenter)) {
/*			printf("# Leaving %s (.onleave)\n", working);*/
			display_file(onenter);
		}
#endif
		while (len && working[len] != '/')
			--len;
		working[len] = 0;
	}

	if (dst[len]) ++len;

	/* Move up to destination */
	while (dst[len] != 0) {
	struct odpath_t *p;

		while (dst[len] != 0 && dst[len] != '/')
			++len;
		strncpy(working, dst, len);
		working[len] = 0;
		p = find_enter(working);
		if (p) {
/*			printf("# Entering %s\n", working);*/
			printf("%s", p->content);
		}
#ifdef USE_ONENTERLEAVE
		snprintf(onenter, PATH_MAX, "%s/.onenter", working);
		if (exec_ok(onenter)) {
/*			printf("# Entering %s (.onenter)\n", working);*/
			display_file(onenter);
		}
#endif
		if (dst[len] == '/') ++len;
	}

	return 0;
}
